# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataset_viber import AnnotatorInterFace

texts = [
    "Anthony Bourdain was an amazing chef!",
    "Anthony Bourdain was a terrible tv persona!",
]
labels = ["positive", "negative"]

interface = AnnotatorInterFace.for_text_classification(
    texts=texts,
    labels=labels,
    dataset_name="username/my_dataset",  # "<my_hf_org>/<my_dataset>" if you want to log to the hub
    hf_token="HF_TOKEN",  # your huggingface token or it will get HF_TOKEN from the environment
    private=False,  # True if you want to keep the dataset private
)
interface.launch()



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect


def _get_init_arg_names(cls) -> list[str]:
    init_signature = inspect.signature(cls.__init__)
    return [param.name for param in init_signature.parameters.values()]


def _get_init_payload(cls) -> dict:
    payload = cls.__dict__
    payload["inputs"] = payload["input_components"]
    payload["outputs"] = payload["output_components"]
    return {
        key: value for key, value in payload.items() if key in _get_init_arg_names(cls)
    }



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import TYPE_CHECKING, Callable, List, Optional, Sequence

import gradio
import huggingface_hub
from gradio.components import Component

from dataset_viber._gradio._flagging import FixedHubDatasetSaver
from dataset_viber._utils import _get_init_payload

if TYPE_CHECKING:
    from transformers.pipelines import Pipeline


class CollectorInterface(gradio.Interface):
    def __init__(
        self,
        fn: Callable,
        inputs: str | Component | Sequence[str | Component] | None,
        outputs: str | Component | Sequence[str | Component] | None,
        *,
        csv_logger: Optional[bool] = False,
        dataset_name: str = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
        allow_flagging: Optional[str] = "auto",
        flagging_options: Optional[List[str]] = None,
        show_embedded_viewer: Optional[bool] = True,
        **kwargs,
    ) -> None:
        """
        Load a CollectorInterface with data logging capabilities.

        Parameters:
            fn: the function to run
            inputs: the input component(s)
            outputs: the output component(s)
            csv_logger: whether or not to log the data to a CSV file.
            dataset_name: the "org/dataset" to which the data needs to be logged
            hf_token: optional token to pass, otherwise will default to env var HF_TOKEN
            private: whether or not to create a private repo
            allow_flagging: One of "never", "auto", or "manual". If "never" or "auto", users will not see a button to flag an input and output. If "manual", users will see a button to flag. If "auto", every input the user submits will be automatically flagged, along with the generated output. If "manual", both the input and outputs are flagged when the user clicks flag button. This parameter can be set with environmental variable GRADIO_ALLOW_FLAGGING; otherwise defaults to "manual".
            flagging_options: If provided, allows user to select from the list of options when flagging. Only applies if allow_flagging is "manual". Can either be a list of tuples of the form (label, value), where label is the string that will be displayed on the button and value is the string that will be stored in the flagging CSV; or it can be a list of strings ["X", "Y"], in which case the values will be the list of strings and the labels will ["Flag as X", "Flag as Y"], etc.

        Return:
            an intialized CollectorInterface
        """
        self.csv_logger = csv_logger
        self._validate_flagging_options(
            allow_flagging=allow_flagging, flagging_options=flagging_options
        )
        flagging_callback = kwargs.pop("flagging_callback", None)
        if dataset_name and flagging_callback is not None:
            flagging_callback = self._get_flagging_callback(
                dataset_name=dataset_name, hf_token=hf_token, private=private
            )

        if csv_logger and not flagging_callback:
            flagging_callback = None
        elif csv_logger and flagging_callback:
            warnings.warn("You are using 2 callbacks, the csv_logger will be ignored.")
        else:
            flagging_callback = gradio.CSVLogger()
            flagging_callback.setup = lambda *args, **kwargs: None
            flagging_callback.flag = lambda *args, **kwargs: 0

        kwargs.update(
            {
                "flagging_callback": flagging_callback,
                "allow_flagging": allow_flagging,
                "flagging_options": flagging_options,
            }
        )
        super().__init__(fn=fn, inputs=inputs, outputs=outputs, **kwargs)
        self = self._add_html_component_with_viewer(
            self, flagging_callback, show_embedded_viewer
        )

    @classmethod
    def from_pipeline(
        cls,
        pipeline: "Pipeline",
        *,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
        allow_flagging: Optional[str] = "auto",
        flagging_options: Optional[List[str]] = None,
        show_embedded_viewer: Optional[bool] = True,
        **kwargs,
    ) -> gradio.Interface:
        """
        Load an existing transformers.pipeline into a CollectorInterface with data logging capabilities.

        Parameters::
            pipeline: an initialized the transformers.pipeline
            csv_logger: whether or not to log the data to a CSV file.
            dataset_name: the "org/dataset" to which the data needs to be logged
            hf_token: optional token to pass, otherwise will default to env var HF_TOKEN
            private: whether or not to create a private repo
            allow_flagging: One of "never", "auto", or "manual". If "never" or "auto", users will not see a button to flag an input and output. If "manual", users will see a button to flag. If "auto", every input the user submits will be automatically flagged, along with the generated output. If "manual", both the input and outputs are flagged when the user clicks flag button. This parameter can be set with environmental variable GRADIO_ALLOW_FLAGGING; otherwise defaults to "manual".
            flagging_options: If provided, allows user to select from the list of options when flagging. Only applies if allow_flagging is "manual". Can either be a list of tuples of the form (label, value), where label is the string that will be displayed on the button and value is the string that will be stored in the flagging CSV; or it can be a list of strings ["X", "Y"], in which case the values will be the list of strings and the labels will ["Flag as X", "Flag as Y"], etc.

        Return:
            an intialized CollectorInterface
        """
        return cls.from_interface(
            interface=gradio.Interface.from_pipeline(pipeline=pipeline),
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
            allow_flagging=allow_flagging,
            flagging_options=flagging_options,
            show_embedded_viewer=show_embedded_viewer,
            **kwargs,
        )

    @classmethod
    def from_interface(
        cls,
        interface: gradio.Interface,
        *,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
        allow_flagging: Optional[str] = "auto",
        flagging_options: Optional[List[str]] = None,
        show_embedded_viewer: Optional[bool] = True,
        **kwargs,
    ) -> gradio.Interface:
        """
        Load an existing gradio.Interface into a CollectorInterface with data logging capabilities.

        Parameters::
            interface: any initialized gradio.Interface
            csv_logger: whether or not to log the data to a CSV file.
            dataset_name: the "org/dataset" to which the data needs to be logged
            hf_token: optional token to pass, otherwise will default to env var HF_TOKEN
            private: whether or not to create a private repo
            allow_flagging: One of "never", "auto", or "manual". If "never" or "auto", users will not see a button to flag an input and output. If "manual", users will see a button to flag. If "auto", every input the user submits will be automatically flagged, along with the generated output. If "manual", both the input and outputs are flagged when the user clicks flag button. This parameter can be set with environmental variable GRADIO_ALLOW_FLAGGING; otherwise defaults to "manual".
            flagging_options: If provided, allows user to select from the list of options when flagging. Only applies if allow_flagging is "manual". Can either be a list of tuples of the form (label, value), where label is the string that will be displayed on the button and value is the string that will be stored in the flagging CSV; or it can be a list of strings ["X", "Y"], in which case the values will be the list of strings and the labels will ["Flag as X", "Flag as Y"], etc.

        Return:
            an intialized CollectorInterface
        """
        flagging_callback = None or kwargs.pop("flagging_callback", None)
        if dataset_name and not flagging_callback:
            flagging_callback = cls._get_flagging_callback(
                dataset_name=dataset_name, hf_token=hf_token, private=private
            )
        payload = _get_init_payload(interface)
        payload.update(**kwargs)
        payload.update(
            {
                "flagging_callback": flagging_callback,
                "allow_flagging": allow_flagging,
                "flagging_options": flagging_options,
                "show_embedded_viewer": show_embedded_viewer,
            }
        )
        return cls(**payload)

    @staticmethod
    def _validate_flagging_options(allow_flagging, flagging_options) -> None:
        if allow_flagging == "auto" and flagging_options:
            raise ValueError(
                "automatic flagging cannot be combined with 'flagging_options', set `allow_flagging='manual'` instead"
            )
        if allow_flagging == "never":
            warnings.warn("You are using a datacollector but don't enable flagging")

    @staticmethod
    def _get_flagging_callback(
        dataset_name: str,
        hf_token: str,
        private: bool = False,
    ) -> gradio.HuggingFaceDatasetSaver:
        return FixedHubDatasetSaver(
            hf_token=hf_token,
            dataset_name=dataset_name,
            private=private,
            info_filename="dataset_info.json",
            separate_dirs=True,
        )

    @staticmethod
    def _get_repo_url_from_repo_id(repo_id: str) -> str:
        return f"https://huggingface.co/datasets/{repo_id}"

    @staticmethod
    def _get_repo_url_fom_dataset_saver(
        flagging_callback: gradio.HuggingFaceDatasetSaver,
    ) -> huggingface_hub.RepoUrl:
        return f"""https://huggingface.co/datasets/{huggingface_hub.create_repo(
            repo_id=flagging_callback.dataset_id,
            token=flagging_callback.hf_token,
            private=flagging_callback.dataset_private,
            repo_type="dataset",
            exist_ok=True,
        ).repo_id}"""

    @staticmethod
    def _get_embedded_dataset_viewer(repo_url: str) -> str:
        return f"""
<iframe
  src="{repo_url}/embed/viewer/default/train"
  frameborder="0"
  width="100%"
  height="560px"
></iframe>
"""

    @classmethod
    def _add_html_component_with_viewer(
        cls,
        instance: gradio.Interface,
        flagging_callback: Optional[gradio.HuggingFaceDatasetSaver] = None,
        show_embedded_viewer: bool = True,
    ) -> gradio.Interface:
        if isinstance(flagging_callback, gradio.HuggingFaceDatasetSaver):
            repo_url = cls._get_repo_url_fom_dataset_saver(flagging_callback)
            formatted_repo_url = (
                f"Data is being written to [a dataset on the Hub]({repo_url})."
            )
            with instance:
                with gradio.Row(equal_height=False):
                    gradio.Markdown(formatted_repo_url)
                if show_embedded_viewer and not flagging_callback.dataset_private:
                    with gradio.Row():
                        with gradio.Accordion(
                            "dataset viewer - do an (empty) search to refresh",
                            open=False,
                        ):
                            gradio.HTML(cls._get_embedded_dataset_viewer(repo_url))
        return instance



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

DEFAULT_EMBEDDING_MODEL = "Snowflake/snowflake-arctic-embed-xs"
COLORS = [
    "#a6cee3",
    "#1f78b4",
    "#b2df8a",
    "#33a02c",
    "#fb9a99",
    "#e31a1c",
    "#fdbf6f",
    "#ff7f00",
    "#cab2d6",
    "#6a3d9a",
    "#ffff99",
    "#b15928",
]



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import sys
from collections import OrderedDict
from pathlib import Path
from typing import Any

import gradio as gr
import huggingface_hub
from gradio import utils
from gradio.flagging import HuggingFaceDatasetSaver
from gradio_client import utils as client_utils

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class FixedHubDatasetSaver(HuggingFaceDatasetSaver):
    @override
    def _deserialize_components(
        self,
        data_dir: Path,
        flag_data: list[Any],
        flag_option: str = "",
        username: str = "",
    ) -> tuple[dict[Any, Any], list[Any]]:
        """Deserialize components and return the corresponding row for the flagged sample.

        Images/audio are saved to disk as individual files.
        """
        # Components that can have a preview on dataset repos
        file_preview_types = {gr.Audio: "Audio", gr.Image: "Image"}

        # Generate the row corresponding to the flagged sample
        features = OrderedDict()
        row = []
        for component, sample in zip(self.components, flag_data):
            # Get deserialized object (will save sample to disk if applicable -file, audio, image,...-)
            label = component.label or ""
            save_dir = data_dir / client_utils.strip_invalid_filename_characters(label)
            save_dir.mkdir(exist_ok=True, parents=True)
            deserialized = utils.simplify_file_data_in_str(
                component.flag(sample, save_dir)
            )

            # Add deserialized object to row
            features[label] = {"dtype": "string", "_type": "Value"}
            try:
                deserialized_path = Path(deserialized)
                if not deserialized_path.exists():
                    raise FileNotFoundError(f"File {deserialized} not found")
                row.append(str(deserialized_path.relative_to(self.dataset_dir)))
            except (FileNotFoundError, TypeError, ValueError, OSError):
                deserialized = "" if deserialized is None else str(deserialized)
                row.append(deserialized)

            # If component is eligible for a preview, add the URL of the file
            # Be mindful that images and audio can be None
            if isinstance(component, tuple(file_preview_types)):  # type: ignore
                for _component, _type in file_preview_types.items():
                    if isinstance(component, _component):
                        features[label + " file"] = {"_type": _type}
                        break
                if deserialized:
                    path_in_repo = str(  # returned filepath is absolute, we want it relative to compute URL
                        Path(deserialized).relative_to(self.dataset_dir)
                    ).replace("\\", "/")
                    row.append(
                        huggingface_hub.hf_hub_url(
                            repo_id=self.dataset_id,
                            filename=path_in_repo,
                            repo_type="dataset",
                        )
                    )
                else:
                    row.append("")
        features["flag"] = {"dtype": "string", "_type": "Value"}
        features["username"] = {"dtype": "string", "_type": "Value"}
        row.append(flag_option)
        row.append(username)
        return features, row



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import os
import threading
import time
import uuid
from pathlib import Path

import gradio
import numpy as np
import pandas as pd
from datasets import Dataset, load_dataset
from gradio_huggingfacehub_search import HuggingfaceHubSearch
from PIL import Image

from dataset_viber._gradio._mixins._argilla import ArgillaMixin

try:
    from gradio.oauth import OAuthToken
    from starlette.middleware.sessions import SessionMiddleware  # noqa
except ImportError:
    OAuthToken = str

CODE_KWARGS = {
    "language": "json",
    "interactive": True,
    "label": "Column Mapping",
    "lines": 1,
}


class ImportExportMixin(ArgillaMixin):
    def _override_block_init_method(self, **kwargs):
        # Initialize the parent class
        gradio.Blocks.__init__(
            self,
            analytics_enabled=kwargs.get("analytics_enabled", True),
            mode="interface",
            css=kwargs.get("css", None),
            title=kwargs.get("title", "Gradio"),
            theme=kwargs.get("theme", None),
            js=kwargs.get("js", None),
            head=kwargs.get("head", None),
            delete_cache=kwargs.get("delete_cache", False),
            fill_width=kwargs.get("fill_width", False),
            # **kwargs,
        )
        # Override the __init__ method of the parent class to avoid the re-creation of the blocks
        gradio.Blocks.__init__ = lambda *args, **kwargs: None

    def _configure_import(self):
        with gradio.Tab("Import data"):
            with gradio.Tab("Import from Hugging Face Hub"):
                search_in = HuggingfaceHubSearch(
                    label="Search Huggingface Hub",
                    placeholder="Search for datasets on Huggingface",
                    search_type="dataset",
                    sumbit_on_select=True,
                )
                dataset_viewer = gradio.HTML(label="Dataset Viewer")
                search_in.submit(
                    fn=lambda x: self._get_embedded_dataset_viewer(
                        self._get_repo_url_from_repo_id(x)
                    ),
                    inputs=[search_in],
                    outputs=[dataset_viewer],
                )
                column_mapping_hf_upload = gradio.Code(
                    value=json.dumps(dict.fromkeys(self.input_columns, ""), indent=2),
                    **CODE_KWARGS,
                )
                start_btn_hf_upload = gradio.Button("Start Annotating")
                start_btn_hf_upload.click(
                    fn=self._set_data_hf_upload,
                    inputs=[search_in, column_mapping_hf_upload],
                    outputs=self.input_data_component,
                )
            with gradio.Tab(label="Import from file"):
                upload_button = gradio.UploadButton(
                    "Upload",
                    label="Select a file (CSV or Excel)",
                    file_types=["csv", "xlsx", "xlsx"],
                )
                df_upload = gradio.Dataframe(interactive=True)
                upload_button.upload(
                    fn=self.upload_file, inputs=upload_button, outputs=df_upload
                )
                column_mapping_file_upload = gradio.Code(
                    value=json.dumps(dict.fromkeys(self.input_columns, ""), indent=2),
                    **CODE_KWARGS,
                )
                start_btn_file_upload = gradio.Button("Start Annotating")
                start_btn_file_upload.click(
                    fn=self._set_data,
                    inputs=[df_upload, column_mapping_file_upload],
                    outputs=self.input_data_component,
                )

    def _configure_export(self):
        with gradio.Tab("Export data"):
            with gradio.Tab("Export to Hugging Face Hub"):
                with gradio.Row():
                    dataset_name = gradio.Textbox(
                        placeholder="Dataset Name", label="Dataset Name"
                    )
                with gradio.Row():
                    gradio.Info(
                        "Ensure HF_TOKEN env var has been set or gradio allows for login through OAuth."
                    )
                    export_button_hf = gradio.Button("Export")
                    try:
                        from starlette.middleware.sessions import SessionMiddleware  # noqa

                        export_button_hf.click(
                            fn=self._export_data_hf,
                            inputs=dataset_name,
                            outputs=dataset_name,
                        )
                    except ImportError:
                        token = gradio.Textbox(
                            value=os.getenv("HF_TOKEN"),
                            type="password",
                            label=f"OAuth Token HF_TOKEN={'HF_TOKEN' in os.environ}",
                            interactive=True,
                        )
                        export_button_hf.click(
                            fn=self._export_data_hf,
                            inputs=[dataset_name, token],
                            outputs=dataset_name,
                        )

            with gradio.Tab("Export to file"):
                with gradio.Column():
                    export_button = gradio.Button("Export")
                self.file = gradio.File(interactive=False, visible=False)
                export_button.click(
                    fn=self._export_data,
                    outputs=self.file,
                )

    def _set_data_hf_upload(self, repo_id, column_mapping, split="train"):
        gradio.Info("Started loading the dataset. This might take a while.")
        try:
            column_mapping = self._json_to_dict(column_mapping)
            dataset = load_dataset(repo_id, split=split)
            for key, value in column_mapping.items():
                if key != value:
                    if value in dataset.column_names:
                        dataset = dataset.rename_column(value, key)
            dataset = dataset.select_columns(
                [
                    col
                    for col in list(column_mapping.keys())
                    if col in dataset.column_names
                ]
            )
            # add images before converting to bytes
            for column in column_mapping.keys():
                if column in dataset.column_names:
                    self.input_data[column].extend(
                        [self.process_image_input(entry) for entry in dataset[column]]
                    )
            self._set_equal_length_input_data()
            dataframe = pd.DataFrame.from_dict(self.input_data)
            self.start = len(dataframe)
        except Exception as e:
            raise gradio.Error(f"An error occurred: {e}")
        gradio.Info(
            "Data loaded successfully. Showing first 100 examples in 'remaing data' tab. Click on \"⏭️ Next\" to get the next record."
        )
        return dataframe.head(100)

    def _set_data(self, dataframe, column_mapping):
        gradio.Info("Started loading the dataset. This might take a while.")
        try:
            column_mapping = self._json_to_dict(column_mapping)
            dataframe = dataframe[list(column_mapping.values())]
            dataframe.columns = list(column_mapping.keys())
            for column in column_mapping.keys():
                if column in dataframe.columns:
                    self.input_data[column].extend(dataframe[column].tolist())
            self._set_equal_length_input_data()
            dataframe = pd.DataFrame.from_dict(self.input_data)
            self.start = len(dataframe)
        except Exception as e:
            raise gradio.Error(f"An error occurred: {e}")
        gradio.Info(
            "Data loaded successfully. Showing first 100 examples in 'remaing data' tab. Click on 🗑️ discard to get the next record."
        )
        return dataframe.head(100)

    def _export_data_hf(self, dataset_name, oauth_token: OAuthToken | None):
        gradio.Info("Started exporting the dataset. This may take a while.")
        if not isinstance(oauth_token, str):
            oauth_token = oauth_token.token
        Dataset.from_dict(self.output_data).push_to_hub(dataset_name, token=oauth_token)
        gradio.Info(f"Exported the dataset to Hugging Face Hub as {dataset_name}.")

    def delete_file_after_delay(self, file_path, delay=30):
        def delete_file():
            time.sleep(delay)
            Path(Path(file_path).name).unlink()

        thread = threading.Thread(target=delete_file)
        thread.start()

    def _export_data(self, dataframe):
        id = uuid.uuid4()
        if "image" in self.task:
            filename = f"{id}.parquet"
            Dataset.from_dict(self.output_data).to_parquet(filename)
        else:
            filename = f"{id}.csv"
            Dataset.from_dict(self.output_data).to_csv(filename)
        self.delete_file_after_delay(filename, 20)
        gradio.Info(
            f"Exported the dataset to {filename}. It will be deleted in 20 seconds."
        )
        return gradio.File(value=filename, visible=True)

    def _delete_file(self, _file):
        return gradio.File(interactive=False, visible=False)

    @staticmethod
    def _json_to_dict(json_str):
        return json.loads(json_str)

    @staticmethod
    def upload_file(file):
        # Determine the file type and load accordingly
        if file.name.endswith(".csv"):
            df = pd.read_csv(file.name)
        elif file.name.endswith(".xls") or file.name.endswith(".xlsx"):
            df = pd.read_excel(file.name)
        else:
            return "Unsupported file type. Please upload a CSV, Excel, or JSON file."
        return df

    def process_image_input(self, input_data):
        if input_data is None:
            return None
        elif isinstance(input_data, dict):
            if "bytes" in input_data and input_data["bytes"]:
                # Case: bytes in a dictionary
                return Image.open(io.BytesIO(input_data["bytes"]))
            elif "path" in input_data and input_data["path"]:
                # Case: path in a dictionary
                return input_data["path"]
        elif isinstance(input_data, Image.Image):
            # Case: PIL Image
            return input_data
        elif isinstance(input_data, str):
            # Case: URL or file path as string
            return input_data
        elif isinstance(input_data, (np.ndarray, list)):
            # Case: numpy array or list
            return Image.fromarray(np.array(input_data))
        else:
            return input_data

    def _set_equal_length_input_data(self):
        # assert all columns for self.input_data are a similar length and fille with "" if not
        max_column_len = max(
            [len(self.input_data[column]) for column in self.input_data.keys()]
        )
        for column in self.input_data.keys():
            if len(self.input_data[column]) < max_column_len:
                self.input_data[column].extend(
                    [""] * (max_column_len - len(self.input_data[column]))
                )

    def _set_equal_length_output_data(self):
        # assert all columns for self.output_data are a similar length and fille with "" if not
        max_column_len = max(
            [len(self.output_data[column]) for column in self.output_data.keys()]
        )
        for column in self.output_data.keys():
            if len(self.output_data[column]) < max_column_len:
                self.output_data[column].extend(
                    [""] * (max_column_len - len(self.output_data[column]))
                )



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import random
import sys
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import gradio
import numpy as np
import pandas as pd
import PIL
import PIL.Image
from gradio.components import (
    Button,
    ClearButton,
)
from gradio.events import Dependency
from gradio.flagging import FlagMethod

from dataset_viber._gradio._mixins._import_export import ImportExportMixin
from dataset_viber._gradio._mixins._task_config import TaskConfigMixin
from dataset_viber._gradio.collector import CollectorInterface

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if TYPE_CHECKING:
    from transformers.pipelines import Pipeline


_POP_INDEX = 0
_MESSAGE_DONE_ANNOTATING = "No data left to annotate."
_HIGHLIGHT_TEXT_KWARGS = {
    "interactive": True,
    "show_legend": True,
    "combine_adjacent": True,
    "adjacent_separator": "",
}
_CHATBOT_KWARGS = {"type": "messages", "label": "prompt", "show_copy_button": True}
_SUBMIT_BTN = gradio.Button("✍🏼 submit", variant="primary", visible=False)
_CLEAR_BTN = gradio.Button("⏭️ Next", variant="stop")
_PREFERENCE_OPTIONS = [
    ("👆 A is better", "A"),
    ("👇 B is better", "B"),
    ("🤝 Tie", "tie"),
]
_SUBMIT_OPTIONS = [("✍🏼 submit", "")]
_NEXT_INPUT_FUNCTION_MESSAGE = (
    "You can only provide `fn_next_input` and not any other values."
)


def wrap_fn_next_input(fn_next_input, output_data, columns):
    def wrapped_fn(*args, **kwargs):
        if args[0]:
            for col, res in zip(columns, args):
                if isinstance(res, np.ndarray):
                    res = PIL.Image.fromarray(res)
                output_data[col].append(res)
        result = fn_next_input(*args, **kwargs)
        return result

    return wrapped_fn


class AnnotatorInterFace(CollectorInterface, ImportExportMixin, TaskConfigMixin):
    @override
    def __init__(
        self,
        *args,
        inputs,
        outputs,
        **kwargs,
    ):
        self._override_block_init_method(**kwargs)
        with self:
            try:
                from starlette.middleware.sessions import SessionMiddleware  # noqa

                gradio.LoginButton(
                    value="Sign in with Hugging Face - a login will reset the data!",
                ).activate()
            except ImportError:
                pass

            with gradio.Accordion(open=False, label="Import and remaining data"):
                with gradio.Tab("Remaining data"):

                    def _get_input_data():
                        self._set_equal_length_input_data()
                        return pd.DataFrame.from_dict(self.input_data).head(100)

                    self.input_data_component = gradio.Dataframe(
                        _get_input_data,
                        interactive=False,
                    )
                    inputs[0].change(
                        fn=_get_input_data,
                        outputs=self.input_data_component,
                    )
                if not list(self.input_data.values())[0]:
                    self._set_text_classification_config(inputs)
                    self._configure_import()
        super().__init__(
            *args,
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )
        with self:
            with gradio.Row():
                with gradio.Column():
                    shuffle = gradio.Button("Shuffle", variant="secondary")
                shuffle.click(
                    fn=self.shuffle_data, inputs=None, outputs=self.input_data_component
                )

            with gradio.Accordion(open=False, label="Export and completed data"):

                def _get_output_data():
                    self._set_equal_length_output_data()
                    return pd.DataFrame.from_dict(self.output_data).tail(100)

                with gradio.Tab("Completed data"):
                    self.output_data_component = gradio.Dataframe(
                        _get_output_data,
                        interactive=False,
                    )
                    inputs[0].change(
                        fn=_get_output_data,
                        outputs=self.output_data_component,
                    )
                self._configure_export()

    @classmethod
    def for_text_classification(
        cls,
        texts: Optional[List[str]] = None,
        suggestions: Optional[List[str]] = None,
        labels: Optional[List[str]] = None,
        multi_label: Optional[bool] = False,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for text classification tasks.

        Parameters:
            texts (Optional[List[str]]): List of texts to annotate.
            suggestions (Optional[List[str]]): List of suggestions to correct for. Defaults to None.
            labels (Optional[List[str]]): List of labels to choose from.
            multi_label (Optional[bool]): Whether to allow multiple labels. Defaults to False.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the text before annotating.
                Expecting it takes a `str` and returns [{"label": str, "score": float}].
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`str`, Union[str, List[str]]].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = (
            "text-classification"
            if not multi_label
            else "text-classification-multi-label"
        )
        cls.labels = labels or []
        cls.input_columns = ["text", "suggestion"]
        cls.output_columns = ["text", "label"]
        cls.input_data = {"text": texts or [], "suggestion": suggestions or []}
        cls.output_data = {label: [] for label in cls.output_columns}
        cls.start = len(cls.input_data["text"])

        # Process function
        if fn_next_input is not None:
            if any(
                [item is not None for item in [texts, suggestions, labels, fn_model]]
            ):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:

            def next_input(
                _text: str, _label: Union[str, List[str]]
            ) -> Tuple[str, Union[str, List[str]]]:
                if _text:
                    cls.output_data["text"].append(_text)
                    cls.output_data["label"].append(_label)
                if cls.input_data["text"]:
                    cls._update_message(cls)
                    text = cls.input_data["text"].pop(_POP_INDEX)
                    label = ""
                    if cls.input_data["suggestion"]:
                        label = cls.input_data["suggestion"].pop(_POP_INDEX)
                    label = label if fn_model is None or label else fn_model(text)
                    if fn_model is not None and not label:
                        label = fn_model(text)
                        if cls.task == "text-classification-multi-label":
                            label = [
                                str(lab["label"]) for lab in label if lab["score"] > 0.5
                            ]
                            return (text, label)
                        return (text, str(label[0]["label"]))
                    else:
                        if cls.task == "text-classification-multi-label":
                            label = [
                                str(lab["label"]) for lab in label if lab["score"] > 0.5
                            ]
                            return (text, label)
                        return (text, str(label))
                else:
                    cls._done_message()
                    return ("", [])

        # UI Config
        text, label = next_input(None, None)
        if cls.task == "text-classification-multi-label":
            check_box_group = gradio.CheckboxGroup(
                cls.labels, value=label, label="label"
            )
        else:
            check_box_group = gradio.Radio(cls.labels, value=label, label="label")
        inputs: List[gradio.Textbox] = [
            gradio.Textbox(value=text, label="text"),
            check_box_group,
        ]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            flagging_options=_SUBMIT_OPTIONS,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_token_classification(
        cls,
        texts: Optional[List[str]] = None,
        labels: Optional[List[str]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for token classification tasks.

        Parameters:
            texts (Optional[List[str]]): List of texts to annotate.
            labels (Optional[List[str]]): List of labels to choose from.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the text before annotating.
                Expecting it takes a `str` and returns List[Tuple[str, str]] [("text","label")].
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`str`,List[Tuple[str, str]]].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "token-classification"
        cls.labels = labels or []
        cls.input_columns = ["text"]
        cls.output_columns = ["text", "spans"]
        cls.input_data = {"text": texts or []}
        cls.output_data = {label: [] for label in cls.output_columns}
        cls.start = len(cls.input_data["text"])

        # Process function
        if fn_next_input is not None:
            if any([item is not None for item in [texts, fn_model]]):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:

            def next_input(
                _text: str, _spans: List[Tuple[str, str]]
            ) -> Tuple[str, List[Tuple[str, str]]]:
                if _text:
                    cls.output_data["text"].append(_text)
                    cls.output_data["spans"].append(_spans)
                if cls.input_data["text"]:
                    cls._update_message(cls)
                    text = cls.input_data["text"].pop(_POP_INDEX)
                    spans = (
                        cls._convert_to_tokens(text)
                        if fn_model is None
                        else fn_model(text)
                    )
                    return text, spans
                else:
                    cls._done_message()
                    return "", cls._convert_to_tokens(" ")

        # UI Config
        text, spans = next_input(None, None)
        inputs = [
            gradio.Textbox(value=text, label="text", interactive=False),
            gradio.HighlightedText(
                value=spans,
                label="spans",
                **_HIGHLIGHT_TEXT_KWARGS,
            ),
        ]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            flagging_options=_SUBMIT_OPTIONS,
            allow_flagging="manual",
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_question_answering(
        cls,
        questions: Optional[List[str]] = None,
        contexts: Optional[List[str]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for question answering tasks.

        Parameters:
            questions (Optional[List[str]]): List of questions to annotate.
            contexts (Optional[List[str]]): List of contexts to annotate.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the text before annotating.
                Expecting it takes a `str` and returns List[Tuple[str, str]] [("text","label")].
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`str`,List[Tuple[str, str]]].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "question-answering"
        cls.input_columns = ["question", "context"]
        cls.output_columns = ["question", "context"]
        cls.input_data = {"question": questions or [], "context": contexts or []}
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["question"])

        # Process function
        if fn_next_input is not None:
            if any([item is not None for item in [questions, contexts, fn_model]]):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            if cls.input_data["question"] and cls.input_data["context"]:
                if len(cls.input_data["question"]) != len(cls.input_data["context"]):
                    raise ValueError(
                        "Questions and contexts must be of the same length."
                    )

            def next_input(
                _question: str, _context: Union[str, List[Tuple[str, str]]]
            ) -> Tuple[str, Union[str, List[Tuple[str, str]]]]:
                if _question:
                    cls.output_data["question"].append(_question)
                    cls.output_data["context"].append(_context)
                if cls.input_data["question"]:
                    cls._update_message(cls)
                    question = cls.input_data["question"].pop(_POP_INDEX)
                    context = cls.input_data["context"].pop(_POP_INDEX)
                    context = (
                        cls._convert_to_tokens(context)
                        if fn_model is None
                        else fn_model(question, context)
                    )
                    return question, context
                else:
                    cls._done_message()
                    return None, []

        # UI Config
        question, context = next_input(None, None)
        input_question = gradio.Textbox(value=question, label="question")
        input_context = gradio.HighlightedText(
            value=context,
            label="context",
            interactive=True,
            show_legend=False,
            combine_adjacent=True,
            adjacent_separator="",
        )
        inputs = [input_question, input_context]
        cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            flagging_options=_SUBMIT_OPTIONS,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_text_generation(
        cls,
        prompts: Optional[List[str]] = None,
        completions: Optional[List[str]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for text generation tasks.

        Parameters:
            prompts (Optional[List[str]]): List of prompts to annotate.
            completions (Optional[List[str]]): List of completions to annotate. Defaults to None.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the prompt before annotating.
                Expecting it takes a `str` and returns `str`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`str`, `str`].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "text-generation"
        cls.input_columns = ["prompt", "completion"]
        cls.output_columns = ["prompt", "completion"]
        cls.input_data = {"prompt": prompts or [], "completion": completions or []}
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["prompt"])

        # Process function
        if fn_next_input is not None:
            if any([item is not None for item in [prompts, completions, fn_model]]):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            if cls.input_data["prompt"] and cls.input_data["completion"]:
                if len(cls.input_data["prompt"]) != len(cls.input_data["completion"]):
                    raise ValueError(
                        "Source and target must be of the same length. You can add empty strings to match the lengths."
                    )

            def next_input(_prompt: str, _completion: str) -> Tuple[str, str]:
                if _prompt:
                    cls.output_data["prompt"].append(_prompt)
                    cls.output_data["completion"].append(_completion)
                if cls.input_data["prompt"]:
                    cls._update_message(cls)
                    prompt = cls.input_data["prompt"].pop(_POP_INDEX)
                    completion = (
                        cls.output_data["completion"].pop(_POP_INDEX)
                        if fn_model is None
                        else fn_model(prompt)
                    )
                    return prompt, completion
                else:
                    cls._done_message()
                    return None, None

        # UI Config
        prompt, completion = next_input(None, None)
        input_prompt = gradio.Textbox(value=prompt, label="prompt")
        input_completion = gradio.Textbox(value=completion, label="completion")
        inputs = [input_prompt, input_completion]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            flagging_options=_SUBMIT_OPTIONS,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_text_generation_preference(
        cls,
        prompts: Optional[List[str]] = None,
        completions_a: Optional[List[str]] = None,
        completions_b: Optional[List[str]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for text generation preference tasks.

        Parameters:
            prompts (Optional[List[str]]): List of prompts to annotate.
            completions_a (Optional[List[str]]): List of completions to annotate for option A. Defaults to None.
            completions_b (Optional[List[str]]): List of completions to annotate for option B. Defaults to None.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the prompt before annotating.
                Expecting it takes a `str` and returns `str`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`str`, `str`, `str`].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "text-generation-preference"
        cls.input_columns = ["prompt", "completion_a", "completion_b"]
        cls.output_columns = ["prompt", "completion_a", "completion_b", "flag"]
        cls.input_data = {
            "prompt": prompts or [],
            "completion_a": completions_a or [],
            "completion_b": completions_b or [],
        }
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["prompt"])

        # Process function
        if fn_next_input is not None:
            if any(
                [item is not None for item in [prompts, completions_a, completions_b]]
            ):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            (
                cls.input_data["prompt"],
                cls.input_data["completion_a"],
                cls.input_data["completion_b"],
            ) = cls._validate_preference(
                fn_model,
                cls.input_data["prompt"],
                cls.input_data["completion_a"],
                cls.input_data["completion_b"],
            )

            def next_input(
                _prompt: str, _completion_a: str, _completion_b: str
            ) -> Tuple[str, str, str]:
                if _prompt:
                    cls.output_data["prompt"].append(_prompt)
                    cls.output_data["completion_a"].append(_completion_a)
                    cls.output_data["completion_b"].append(_completion_b)
                if cls.input_data["prompt"]:
                    cls._update_message(cls)
                    prompt = cls.input_data["prompt"].pop(_POP_INDEX)
                    completion_a = cls.input_data["completion_a"].pop(_POP_INDEX)
                    completion_b = cls.input_data["completion_b"].pop(_POP_INDEX)
                    completion_a = (
                        completion_a
                        if fn_model is None or completion_a != ""
                        else fn_model(prompt)
                    )
                    completion_b = (
                        completion_b
                        if fn_model is None or completion_b != ""
                        else fn_model(prompt)
                    )
                    return prompt, completion_a, completion_b
                else:
                    cls._done_message()
                    return None, None, None

        # UI Config
        prompt, completion_a, completion_b = next_input(None, None, None)
        input_prompt = gradio.Textbox(value=prompt, label="prompt")
        input_completion_a = gradio.Textbox(value=completion_a, label="👆 completion A")
        input_completion_b = gradio.Textbox(value=completion_b, label="👇 completion B")
        inputs = [input_prompt, input_completion_a, input_completion_b]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            flagging_options=_PREFERENCE_OPTIONS,
            clear_btn=_CLEAR_BTN,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_chat_classification(
        cls,
        prompts: Optional[List[List[Dict[str, str]]]] = None,
        suggestions: Optional[List[str]] = None,
        labels: Optional[List[str]] = None,
        multi_label: Optional[bool] = False,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for chat classification tasks.

        Parameters:
            prompts (Optional[List[List[Dict[str, str]]]): List of chat messages to annotate.
            suggestions (Optional[List[str]]): List of suggestions to correct for. Defaults to None.
            labels (Optional[List[str]]): List of labels to choose from.
            multi_label (Optional[bool]): Whether to allow multiple labels. Defaults to False.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the chat messages before annotating.
                Expecting it takes a `str` and returns [{"label": str, "score": float}].
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`List[gradio.ChatMessage]`, List[str]].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = (
            "chat-classification"
            if not multi_label
            else "chat-classification-multi-label"
        )
        cls.labels = labels
        cls.input_columns = ["prompt", "suggestion"]
        cls.output_columns = ["prompt", "label"]
        cls.input_data = {"prompt": prompts or [], "suggestion": suggestions or []}
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["prompt"])

        # Process function
        if fn_next_input is not None:
            if any([item is not None for item in [prompts, suggestions, fn_model]]):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            cls.input_data["prompt"] = cls._convert_to_chat_message(
                cls.input_data["prompt"]
            )

            def next_input(
                _prompt: List[gradio.ChatMessage], _label: Union[str, List[str]]
            ) -> Tuple[List[gradio.ChatMessage], Union[str, List[str]]]:
                if _prompt:
                    cls.output_data["prompt"].append(_prompt)
                    cls.output_data["label"].append(_label)
                if cls.input_data["prompt"]:
                    cls._update_message(cls)
                    prompt = cls.input_data["prompt"].pop(_POP_INDEX)
                    label = ""
                    if cls.input_data["suggestion"]:
                        label = cls.input_data["suggestion"].pop(_POP_INDEX)
                    if fn_model is not None and not label:
                        label = fn_model("\n".join([msg.content for msg in prompt]))
                        if cls.task == "chat-classification-multi-label":
                            label = [
                                str(lab["label"]) for lab in label if lab["score"] > 0.5
                            ]
                            return (prompt, label)
                        return (prompt, str(label[0]["label"]))
                    else:
                        if cls.task == "chat-classification-multi-label":
                            label = [str(lab) for lab in label]
                            return (prompt, label)
                        return (prompt, str(label))
                else:
                    cls._done_message()
                    return (None, [])

        # UI Config
        prompt, label = next_input(None, None)
        if cls.task == "chat-classification-multi-label":
            check_box_group = gradio.CheckboxGroup(
                cls.labels, value=label, label="label"
            )
        else:
            check_box_group = gradio.Radio(cls.labels, value=label, label="label")
        inputs = [
            gradio.Chatbot(
                value=prompt,
                **_CHATBOT_KWARGS,
            ),
            check_box_group,
        ]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            flagging_options=_SUBMIT_OPTIONS,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_chat_generation(
        cls,
        prompts: Optional[
            Union[List[List[Dict[str, str]]], List[List[gradio.ChatMessage]]]
        ] = None,
        completions: Optional[List[str]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for chat generation tasks.

        Parameters:
            prompts (Optional[Union[List[List[Dict[str, str]]], List[List[gradio.ChatMessage]]]): List of chat messages to annotate.
            completions (Optional[List[str]]): List of completions to annotate. Defaults to None.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the chat messages before annotating.
                Expecting it takes a `List[gradio.ChatMessage]` and returns `str`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`List[gradio.ChatMessage]`, `str`].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "chat-generation"
        cls.input_columns = ["prompt", "completion"]
        cls.output_columns = ["prompt", "completion"]
        cls.input_data = {"prompt": prompts or [], "completion": completions or []}
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["prompt"])

        # Process function
        if fn_next_input is not None:
            if any([item is not None for item in [prompts, completions, fn_model]]):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            if cls.input_data["prompt"] and cls.input_data["completion"]:
                if len(cls.input_data["prompt"]) != len(cls.input_data["completion"]):
                    raise ValueError(
                        "Source and target must be of the same length. You can add empty strings to match the lengths."
                    )
            cls.input_data["prompt"] = cls._convert_to_chat_message(
                cls.input_data["prompt"]
            )

            def next_input(_prompt: str, _completion: str) -> Tuple[str, str]:
                def _last_is_user(_prompt):
                    return _prompt[-1].role == "user"

                if _prompt:
                    cls.output_data["prompt"].append(_prompt)
                    cls.output_data["completion"].append(_completion)
                if cls.input_data["prompt"]:
                    cls._update_message(cls)
                    prompt = cls.input_data["prompt"].pop(_POP_INDEX)
                    completion = cls.input_data["completion"].pop(_POP_INDEX)
                    completion = (
                        completion
                        if (fn_model is None or completion != "")
                        and _last_is_user(prompt)
                        else fn_model(prompt)
                    )
                    return prompt, completion
                else:
                    cls._done_message()
                    return [], None

        # UI Config
        prompt, completion = next_input(None, None)
        input_prompt = gradio.Chatbot(
            value=prompt,
            **_CHATBOT_KWARGS,
        )
        input_completion = gradio.Textbox(value=completion, label="completion")
        inputs = [input_prompt, input_completion]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            flagging_options=_SUBMIT_OPTIONS,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_chat_generation_preference(
        cls,
        prompts: Optional[
            Union[List[List[Dict[str, str]]], List[List[gradio.ChatMessage]]]
        ] = None,
        completions_a: Optional[List[str]] = None,
        completions_b: Optional[List[str]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for chat generation preference tasks.

        Parameters:
            prompts (Optional[Union[List[List[Dict[str, str]]], List[List[gradio.ChatMessage]]]): List of chat messages to annotate.
            completions_a (Optional[List[str]]): List of completions to annotate for option A.
            completions_b (Optional[List[str]]): List of completions to annotate for option B.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the chat messages before annotating.
                Expecting it takes a `List[gradio.ChatMessage]` and returns `str`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`List[gradio.ChatMessage]`, `str`, `str`].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "chat-generation-preference"
        cls.input_columns = ["prompt", "completion_a", "completion_a"]
        cls.output_columns = ["prompt", "completion_a", "completion_a", "flag"]
        cls.input_data = {
            "prompt": prompts or [],
            "completion_a": completions_a or [],
            "completion_b": completions_b or [],
        }
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["prompt"])

        # Process function
        if fn_next_input is not None:
            if any(
                [item is not None for item in [prompts, completions_a, completions_b]]
            ):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            (
                cls.input_data["prompt"],
                cls.input_data["completion_a"],
                cls.input_data["completion_b"],
            ) = cls._validate_preference(
                fn_model,
                cls.input_data["prompt"],
                cls.input_data["completion_a"],
                cls.input_data["completion_b"],
            )
            cls.input_data["prompt"] = cls._convert_to_chat_message(
                cls.input_data["prompt"]
            )

            def next_input(
                _prompt: List[gradio.ChatMessage],
                _completion_a: str,
                _completion_b: str,
            ) -> Tuple[List[gradio.ChatMessage], str, str]:
                if _prompt:
                    cls.output_data["prompt"].append(_prompt)
                    cls.output_data["completion_a"].append(_completion_a)
                    cls.output_data["completion_b"].append(_completion_b)
                if cls.input_data["prompt"]:
                    cls._update_message(cls)
                    prompt = cls.input_data["prompt"].pop(_POP_INDEX)
                    completion_a = cls.input_data["completion_a"].pop(_POP_INDEX)
                    completion_b = cls.input_data["completion_b"].pop(_POP_INDEX)
                    completion_a = (
                        completion_a
                        if fn_model is None or completion_a != ""
                        else fn_model(prompt)
                    )
                    completion_b = (
                        completion_b
                        if fn_model is None or completion_b != ""
                        else fn_model(prompt)
                    )
                    return prompt, completion_a, completion_b
                else:
                    cls._done_message()
                    return [], None, None

        # UI Config
        prompt, completion_a, completion_b = next_input(None, None, None)
        input_prompt = gradio.Chatbot(value=prompt, **_CHATBOT_KWARGS)
        input_completion_a = gradio.Textbox(value=completion_a, label="👆 completion A")
        input_completion_b = gradio.Textbox(value=completion_b, label="👇 completion B")
        inputs = [input_prompt, input_completion_a, input_completion_b]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            flagging_options=_PREFERENCE_OPTIONS,
            clear_btn=_CLEAR_BTN,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_image_classification(
        cls,
        images: Optional[List[Union[np.array, PIL.Image.Image, str]]] = None,
        suggestions: Optional[List[str]] = None,
        labels: Optional[List[str]] = None,
        multi_label: Optional[bool] = False,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for image classification tasks.

        Parameters:
            images (List[Union[np.array, PIL.Image.Image, str]]): List of images to annotate.
            suggestions (Optional[List[str]]): List of suggestions to correct for. Defaults to None.
            labels (List[str]): List of labels to choose from.
            multi_label (Optional[bool]): Whether to allow multiple labels. Defaults to False.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the image before annotating.
                it should take an `np.array` or `PIL.Image.Image` and return `Union[str, List[Dict[str, Union[str, float]]]]`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[PIL.Image.Image, Union[str, List[str]]].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = (
            "image-classification"
            if not multi_label
            else "image-classification-multi-label"
        )
        cls.labels = labels or []
        cls.input_columns = ["image", "suggestion"]
        cls.output_columns = ["image", "label"]
        cls.input_data = {"image": images or [], "suggestion": suggestions or []}
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["image"])

        # Process function
        if fn_next_input is not None:
            if any([item is not None for item in [images, suggestions, fn_model]]):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:

            def next_input(
                _image: PIL.Image.Image, _label: str
            ) -> Tuple[PIL.Image.Image, str]:
                if _image is not None:
                    _image = cls.process_image_input(cls, _image)
                    cls.output_data["image"].append(_image)
                    cls.output_data["label"].append(_label)
                if cls.input_data["image"]:
                    cls._update_message(cls)
                    image = cls.input_data["image"].pop(_POP_INDEX)
                    label = ""
                    if cls.input_data["suggestion"]:
                        label = cls.input_data["suggestion"].pop(_POP_INDEX)
                    if fn_model is not None and not label:
                        label = fn_model(image)
                        if cls.task == "image-classification-multi-label":
                            label = [
                                str(lab["label"]) for lab in label if lab["score"] > 0.5
                            ]
                            return (image, label)
                        return (image, str(label[0]["label"]))
                    else:
                        if cls.task == "image-classification-multi-label":
                            label = [str(lab) for lab in label]
                        return (image, str(label))
                else:
                    cls._done_message()
                    return (None, [])

        # UI Config
        image, label = next_input(None, None)
        inputs = gradio.Image(value=image, label="image", height=400)
        if cls.task == "image-classification-multi-label":
            check_box_group = gradio.CheckboxGroup(
                cls.labels, value=label, label="label"
            )
        else:
            check_box_group = gradio.Radio(cls.labels, value=label, label="label")
        inputs = [inputs, check_box_group]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            flagging_options=_SUBMIT_OPTIONS,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_image_generation(
        cls,
        prompts: Optional[List[str]] = None,
        completions: Optional[List[Union[np.array, PIL.Image.Image, str]]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for image generation tasks.

        Parameters:
            prompts (Optional[List[str]]): List of prompts to annotate.
            completions (Optional[List[Union[np.array, PIL.Image.Image, str]]]): List of completions to annotate. Defaults to None.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the prompt to generate an image.
                it takes a `str` and returns `Union[np.array, PIL.Image.Image, str]`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`str`, PIL.Image.Image]`.
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "image-generation"
        cls.input_columns = ["prompt", "completion"]
        cls.output_columns = ["prompt", "completion"]
        cls.input_data = {"prompt": prompts or [], "completion": completions or []}
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start: int = len(cls.input_data["prompt"])

        # Process function
        if fn_next_input is not None:
            if any([item is not None for item in [prompts, completions, fn_model]]):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )

        # Input validation
        if cls.input_data["prompt"] and cls.input_data["completion"]:
            if len(cls.input_data["prompt"]) != len(cls.input_data["completion"]):
                raise ValueError(
                    "Source and target must be of the same length. You can add empty strings to match the lengths."
                )
        else:

            def next_input(
                _prompt: str, _completion: PIL.Image.Image
            ) -> Tuple[str, PIL.Image.Image]:
                if _prompt:
                    _completion = cls.process_image_input(cls, _completion)
                    cls.output_data["prompt"].append(_prompt)
                    cls.output_data["completion"].append(_completion)
                if cls.input_data["prompt"]:
                    cls._update_message(cls)
                    prompt = cls.input_data["prompt"].pop(_POP_INDEX)
                    completion = cls.input_data["completion"].pop(_POP_INDEX)
                    completion = (
                        completion
                        if fn_model is None or completion
                        else fn_model(prompt)
                    )
                    return prompt, completion
                else:
                    cls._done_message()
                    return None, None

        # UI Config
        prompt, completion = next_input(None, None)
        input_prompt = gradio.Textbox(value=prompt, label="prompt")
        input_completion = gradio.Image(
            value=completion, label="completion", height=400
        )
        inputs = [input_prompt, input_completion]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            flagging_options=_SUBMIT_OPTIONS,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_image_description(
        cls,
        images: Optional[List[Union[np.array, PIL.Image.Image, str]]] = None,
        descriptions: Optional[List[str]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for image description tasks.

        Parameters:
            images (List[Union[np.array, PIL.Image.Image, str]]): List of images to annotate.
            descriptions (Optional[List[str]]): List of descriptions to annotate. Defaults to None.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the image before annotating.
                it should take an `np.array` or `PIL.Image.Image` and return `str`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[PIL.Image.Image, str].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "image-description"
        cls.input_columns = ["image", "description"]
        cls.output_columns = ["image", "description"]
        cls.input_data = {"image": images or [], "description": descriptions or []}
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["image"])

        # Process function
        if fn_next_input is not None:
            if any([item is not None for item in [images, descriptions, fn_model]]):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            if cls.input_data["prompt"] and cls.input_data["completion"]:
                if len(cls.input_data["prompt"]) != len(cls.input_data["completion"]):
                    raise ValueError(
                        "Source and target must be of the same length. You can add empty strings to match the lengths."
                    )

            def next_input(
                _image: PIL.Image.Image, _description: str
            ) -> Tuple[PIL.Image.Image, str]:
                if _image is not None:
                    _image = cls.process_image_input(cls, _image)
                    cls.output_data["image"].append(_image)
                    cls.output_data["description"].append(_description)
                if cls.input_data["image"]:
                    cls._update_message(cls)
                    image = cls.input_data["image"].pop(_POP_INDEX)
                    description = cls.input_data["description"].pop(_POP_INDEX)
                    description = (
                        description
                        if fn_model is None or description
                        else fn_model(image)
                    )
                    return image, description
                else:
                    cls._done_message()
                    return None, ""

        # UI Config
        image, description = next_input(None, None)
        inputs = gradio.Image(value=image, label="image", height=400)
        outputs = gradio.Textbox(value=description, label="text")
        inputs = [inputs, outputs]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            flagging_options=_SUBMIT_OPTIONS,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_image_generation_preference(
        cls,
        prompts: Optional[List[str]] = None,
        completions_a: Optional[List[Union[np.array, PIL.Image.Image, str]]] = None,
        completions_b: Optional[List[Union[np.array, PIL.Image.Image, str]]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for image generation preference tasks.

        Parameters:
            prompts (List[str]): List of prompts to annotate.
            completions_a (List[Union[np.array, PIL.Image.Image, str]]): List of completions to annotate for option A.
            completions_b (List[Union[np.array, PIL.Image.Image, str]]): List of completions to annotate for option B.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the prompt to generate an image.
                it should take a `str` and return `PIL.Image.Image`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[`str`, PIL.Image.Image, PIL.Image.Image].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.
        """
        # IO Config
        cls.task = "image-generation-preference"
        cls.input_columns = ["prompt", "completion_a", "completion_b"]
        cls.output_columns = ["prompt", "completion_a", "completion_b", "flag"]
        cls.input_data = {
            "prompt": prompts or [],
            "completion_a": completions_a or [],
            "completion_b": completions_b or [],
        }
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["prompt"])

        # Process function
        if fn_next_input is not None:
            if any(
                [item is not None for item in [prompts, completions_a, completions_b]]
            ):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            if (
                cls.input_data["prompt"]
                and cls.input_data["completion_a"]
                and cls.input_data["completion_b"]
            ):
                if (
                    len(cls.input_data["prompt"])
                    != len(cls.input_data["completion_a"])
                    != len(cls.input_data["completion_b"])
                ):
                    raise ValueError(
                        "Prompts and completions must be of the same length"
                    )

            def next_input(
                _prompt: str,
                _completion_a: PIL.Image.Image,
                _completion_b: PIL.Image.Image,
            ) -> Tuple[str, PIL.Image.Image, PIL.Image.Image]:
                if _prompt:
                    _completion_a = cls.process_image_input(cls, _completion_a)
                    _completion_b = cls.process_image_input(cls, _completion_b)
                    cls.output_data["completion_a"].append(_completion_a)
                    cls.output_data["completion_b"].append(_completion_b)
                if cls.input_data["prompt"]:
                    cls._update_message(cls)
                    prompt = cls.input_data["prompt"].pop(_POP_INDEX)
                    completion_a = cls.input_data["completion_a"].pop(_POP_INDEX)
                    completion_b = cls.input_data["completion_b"].pop(_POP_INDEX)
                    completion_a = (
                        completion_a
                        if fn_model is None or completion_a
                        else fn_model(prompt)
                    )
                    completion_b = (
                        completion_b
                        if fn_model is None or completion_b
                        else fn_model(prompt)
                    )
                    return prompt, completion_a, completion_b
                else:
                    cls._done_message()
                    return None, None, None

        # UI Config
        prompt, completion_a, completion_b = next_input(None, None, None)
        input_prompt = gradio.Textbox(value=prompt, label="prompt")
        input_completion_a = gradio.Image(
            value=completion_a, label="👆 completion A", height=400
        )
        input_completion_b = gradio.Image(
            value=completion_b, label="👇 completion B", height=400
        )
        inputs = [input_prompt, input_completion_a, input_completion_b]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            flagging_options=_PREFERENCE_OPTIONS,
            clear_btn=_CLEAR_BTN,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @classmethod
    def for_image_question_answering(
        cls,
        images: List[Union[np.array, PIL.Image.Image, str]] = None,
        questions: Optional[List[str]] = None,
        answers: Optional[List[str]] = None,
        interactive: Optional[Union[bool, List[bool]]] = True,
        fn_model: Optional[Union["Pipeline", callable]] = None,
        fn_next_input: Optional[callable] = None,
        csv_logger: Optional[bool] = False,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ) -> "AnnotatorInterFace":
        """
        Annotator Interface for image question answering tasks.

        Parameters:
            images (List[Union[np.array, PIL.Image.Image, str]]): List of images to annotate.
            questions (Optional[List[str]]): List of questions to annotate. Defaults to None.
            answers (Optional[List[str]]): List of answers to annotate. Defaults to None.
            interactive (Optional[Union[bool, List[bool]]]): Whether to allow interactive gradio components.
                If a list is provided, it should be of the same length as the number of inputs.
                Defaults to True.
            fn_model (Optional[Union["Pipeline", callable]]): Prediction function to apply to the image before annotating.
                it should take an `PIL.Image.Image` and `str` and return `str`.
                Defaults to None.
            fn_next_input (Optional[callable]): Custom function to forward the next input in an active setting.
                Expecting it takes and return a Tuple[PIL.Image.Image, str, str].
                Defaults to None.
            csv_logger (Optional[bool]): Whether to log the annotations to a CSV file. Defaults to False.
            dataset_name (Optional[str]): Name of the dataset to save the annotations. Defaults to None.
            hf_token (Optional[str]): Hugging Face API token to save the annotations. Defaults to None.
            private (Optional[bool]): Whether to save the annotations as private. Defaults to False.

        Returns:
            AnnotatorInterFace: An instance of AnnotatorInterFace
        """
        # IO Config
        cls.task = "image-question-answering"
        cls.input_columns = ["image", "question", "answer"]
        cls.output_columns = ["image", "question", "answer"]
        cls.input_data = {
            "image": images or [],
            "question": questions or [],
            "answer": answers or [],
        }
        cls.output_data = {col: [] for col in cls.output_columns}
        cls.start = len(cls.input_data["image"])

        # Process function
        if fn_next_input is not None:
            if any(
                [item is not None for item in [images, questions, answers, fn_model]]
            ):
                raise ValueError(_NEXT_INPUT_FUNCTION_MESSAGE)
            next_input = wrap_fn_next_input(
                fn_next_input, cls.output_data, cls.output_columns
            )
        else:
            # Input validation
            if cls.input_data["image"]:
                max_length = max(len(img) for img in cls.input_data["image"])
                cls.input_data["question"] = [
                    question.ljust(max_length)
                    for question in cls.input_data["question"]
                ]
                cls.input_data["answer"] = [
                    answer.ljust(max_length) for answer in cls.input_data["answer"]
                ]

            def next_input(_image, _question, _answer):
                if _image is not None:
                    _image = cls.process_image_input(cls, _image)
                    cls.output_data["image"].append(_image)
                    cls.output_data["question"].append(_question)
                    cls.output_data["answer"].append(_answer)
                if cls.input_data["image"]:
                    cls._update_message(cls)
                    image = cls.input_data["image"].pop(_POP_INDEX)
                    question = cls.input_data["question"].pop(_POP_INDEX)
                    answer = cls.input_data["answer"].pop(_POP_INDEX)
                    if image and question:
                        if fn_model is not None and not answer:
                            answer = fn_model(image, question)
                    return image, question, answer
                else:
                    cls._done_message()
                    return None, "", ""

        # UI Config
        image, question, answer = next_input(None, None, None)
        inputs = [
            gradio.Image(value=image, label="image", height=400),
            gradio.Textbox(value=question, label="question"),
            gradio.Textbox(value=answer, label="answer"),
        ]
        inputs = cls._validate_and_assign_interactive_inputs(inputs, interactive)
        return cls(
            fn=next_input,
            inputs=inputs,
            outputs=inputs,
            flagging_options=_SUBMIT_OPTIONS,
            allow_flagging="manual",
            submit_btn=_SUBMIT_BTN,
            clear_btn=_CLEAR_BTN,
            csv_logger=csv_logger,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
        )

    @override
    def attach_flagging_events(
        self,
        flag_btns: list[Button, gradio.CheckboxGroup] | None,
        _clear_btn: ClearButton,
        _submit_event: Dependency,
    ) -> None:
        """Override the attach_flagging_events method to attach the flagging events."""
        # before the flagging because otherwise input is reset
        self.attach_submit_events(_submit_btn=_clear_btn, _stop_btn=None)

        def add_label(value):
            if "flag" in self.output_data:
                self.output_data["flag"].append(value)

        for btn in flag_btns:
            partial = functools.partial(add_label, btn.value)
            btn.click(fn=partial)
        super().attach_flagging_events(flag_btns, _clear_btn, _submit_event)
        if self.allow_flagging == "manual":
            for flag_btn in flag_btns:
                if flag_btn.label != "✍🏼 submit":
                    self.attach_submit_events(_submit_btn=flag_btn, _stop_btn=None)
                if flag_btn.label == "✍🏼 submit":
                    flag_method = FlagMethod(
                        self.flagging_callback, "", "", visual_feedback=False
                    )
                    flag_btn.click(
                        flag_method,
                        inputs=self.input_components + self.output_components,
                        outputs=None,
                        preprocess=False,
                        queue=False,
                        show_api=False,
                    )

    @override
    def render_flag_btns(self) -> list[Button]:
        """Override the render_flag_btns method to return the flagging buttons with labels."""
        self.flagging_btns = [
            Button(label, variant="primary") for label, _ in self.flagging_options
        ]
        return self.flagging_btns

    def _update_message(self) -> None:
        """Print the progress of the annotation."""
        key = list(self.input_data.keys())[0]
        items = self.input_data[key]
        start = len(self.input_data[key]) + len(self.output_data[key])
        gradio.Info(f"{(len(items) / start) * 100:.2f}% done {len(items)} left.")

    @staticmethod
    def _done_message() -> None:
        """Print the done message."""
        gradio.Info(_MESSAGE_DONE_ANNOTATING)

    @staticmethod
    def _convert_to_tokens(text: str) -> List[tuple[str, None]]:
        """Convert a string to a list of tokens for TextHighligh."""
        return [(char, None) for char in text]

    @staticmethod
    def _convert_to_chat_message(
        messages: Union[List[List[Dict[str, str]]], List[List[gradio.ChatMessage]]],
        with_turn=False,
        last_role=None,
    ) -> List[List[gradio.ChatMessage]]:
        """
        Convert a list of chat messages to a list of gradio.ChatMessage.

        Parameters:
            messages (Union[List[List[Dict[str, str]]], List[List[gradio.ChatMessage]]): List of chat messages.
            with_turn (bool): Whether to add turn information. Defaults to False.
            last_role ([type]): Last role. Defaults to None.

        Returns:
            List[List[gradio.ChatMessage]]: List of chat messages.
        """
        if not messages:
            return []
        if isinstance(messages[0][0], dict):
            messages = [
                [gradio.ChatMessage(**msg) for msg in prompt] for prompt in messages
            ]
        elif isinstance(messages[0][0], str):
            messages = [
                [gradio.ChatMessage(role="user", content=msg) for msg in messages]
            ]
        elif isinstance(messages[0][0], gradio.ChatMessage):
            pass
        else:
            raise ValueError(
                "messages should be a list of list of dict or list of list of gradio.ChatMessage"
            )

        if with_turn:
            messages = [
                [
                    gradio.ChatMessage(
                        role=msg.role,
                        content=msg.content,
                        metadata={
                            "title": f"Turn {idx} - role: {msg.role} - length {len(msg.content)}"
                        },
                    )
                    for idx, msg in enumerate(prompt)
                ]
                for prompt in messages
            ]
        if last_role is not None:
            for prompt in messages:
                assert (
                    prompt[-1].role == last_role
                ), f"Last message role should be {last_role}."
        return messages

    @staticmethod
    def _validate_preference(fn, prompts, completions_a, completions_b):
        """
        Validate the inputs for preference tasks.

        Parameters:
            fn: Prediction function.
            prompts: List of prompts.
            completions_a: List of completions for option A.
            completions_b: List of completions for option B.

        Returns:
            Tuple: Tuple of prompts, completions_a, completions_b.
        """
        if fn is not None and (completions_a is not None and completions_b is not None):
            raise ValueError("fn should be None when completions are provided.")
        if completions_a is None:
            completions_a = ["" for _ in range(len(prompts))]
        if completions_b is None:
            completions_b = ["" for _ in range(len(prompts))]
        if any(
            [len(prompts) != len(completions_a), len(prompts) != len(completions_b)]
        ):
            raise ValueError("Prompts and completions must be of the same length")
        return prompts, completions_a, completions_b

    @staticmethod
    def _validate_and_assign_interactive_inputs(
        inputs: List[gradio.component], interactive: Union[bool, List[bool]]
    ) -> List[gradio.component]:
        """
        Validate and assign the interactive inputs.

        Parameters:
            inputs (List[gradio.component]): List of inputs.
            interactive (Union[bool, List[bool]]): Interactive inputs.

        Returns:
            List[gradio.component]: List of inputs.
        """
        if isinstance(interactive, bool):
            interactive = [interactive] * len(inputs)
        if len(inputs) != len(interactive):
            raise ValueError(
                f"inputs and interactive must be of the same length. Interactive is of length {len(interactive)} but should be {len(inputs)}."
            )
        for input_, is_interactive in zip(inputs, interactive):
            input_.interactive = is_interactive
        return inputs

    def shuffle_data(self):
        if not self.input_data.values():
            return self.input_data

        # Get the length of the first list in the dictionary
        first_key = next(iter(self.input_data))
        length = len(self.input_data[first_key])

        # Check if all lists have the same length
        if not all(len(lst) == length for lst in self.input_data.values()):
            raise ValueError("All input lists must have the same length")

        # Create a list of indices and shuffle it
        indices = list(range(length))
        random.shuffle(indices)

        # Reorder each list based on the shuffled indices
        gradio.Info("Data shuffled.")
        self.input_data = {
            key: [lst[i] for i in indices] for key, lst in self.input_data.items()
        }
        return pd.DataFrame.from_dict(self.input_data).head(100)



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import torch
import torch.nn.functional as F
from optimum.modeling_base import OptimizedModel
from optimum.onnxruntime import ORTModelForFeatureExtraction, ORTOptimizer, ORTQuantizer
from optimum.onnxruntime.configuration import AutoQuantizationConfig, OptimizationConfig
from transformers import AutoTokenizer, Pipeline

from dataset_viber._constants import DEFAULT_EMBEDDING_MODEL


class Embedder:
    def __init__(
        self,
        model_id=DEFAULT_EMBEDDING_MODEL,
        use_onnx=True,
        device="cpu",
    ):
        self.device = device
        self.model_id = model_id
        # set onnx path to cache dir in main home directory
        self.onnx_path = Path("~/.cache/onnx").expanduser()
        self.use_onnx = use_onnx
        self.model = None
        self.tokenizer = None
        self.pipeline = None
        self.load_model()
        self.save_model()
        self.create_pipeline()
        print(f"Model loaded on {self.device}")
        if self.use_onnx:
            print("Optimizing and quantizing model...")
            self.optimize_model()
            print("Optimization complete.")
            self.load_optimized_model()
            self.quantize_model()
            self.print_model_sizes()

    def load_model(self):
        if self.use_onnx:
            self.model: OptimizedModel = ORTModelForFeatureExtraction.from_pretrained(
                self.model_id, export=True
            )
        else:
            self.model = ORTModelForFeatureExtraction.from_pretrained(
                self.model_id, device=self.device
            )
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)

    def save_model(self):
        self.model.save_pretrained(self.onnx_path)
        self.tokenizer.save_pretrained(self.onnx_path)

    @staticmethod
    def mean_pooling(model_output, attention_mask):
        token_embeddings = model_output[0]
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
            input_mask_expanded.sum(1), min=1e-9
        )

    class SentenceEmbeddingPipeline(Pipeline):
        def _sanitize_parameters(self, **kwargs):
            return {}, {}, {}

        def preprocess(self, inputs):
            return self.tokenizer(
                inputs, padding=True, truncation=True, return_tensors="pt"
            )

        def _forward(self, model_inputs):
            outputs = self.model(**model_inputs)
            return {
                "outputs": outputs,
                "attention_mask": model_inputs["attention_mask"],
            }

        def postprocess(self, model_outputs):
            sentence_embeddings = Embedder.mean_pooling(
                model_outputs["outputs"], model_outputs["attention_mask"]
            )
            return F.normalize(sentence_embeddings, p=2, dim=1)

    def create_pipeline(self):
        self.pipeline = self.SentenceEmbeddingPipeline(
            model=self.model, tokenizer=self.tokenizer, device=self.device
        )

    def encode(self, text, convert_to_numpy=False):
        prediction = self.pipeline(text)
        if convert_to_numpy:
            return [pred.cpu().detach().numpy()[0] for pred in prediction]
        else:
            return prediction

    def optimize_model(self):
        if not self.use_onnx:
            raise ValueError(
                "Model must be in ONNX format to optimize. Set use_onnx=True when initializing."
            )
        optimizer = ORTOptimizer.from_pretrained(self.model)
        optimization_config = OptimizationConfig(
            optimization_level=99,
            fp16=True,
            optimize_for_gpu=False if self.device == "cpu" else True,
        )
        optimizer.optimize(
            save_dir=self.onnx_path,
            file_suffix=f"optimized_{self.device}",
            optimization_config=optimization_config,
        )

    def load_optimized_model(self):
        if not self.use_onnx:
            raise ValueError(
                "Model must be in ONNX format to load optimized version. Set use_onnx=True when initializing."
            )
        self.model = ORTModelForFeatureExtraction.from_pretrained(
            self.onnx_path,
            file_name=f"model_optimized_{self.device}.onnx",
            device=self.device,
        )
        self.create_pipeline()

    def quantize_model(self):
        if not self.use_onnx:
            raise ValueError(
                "Model must be in ONNX format to quantize. Set use_onnx=True when initializing."
            )
        dynamic_quantizer = ORTQuantizer.from_pretrained(self.model)
        dqconfig = AutoQuantizationConfig.avx512_vnni(
            is_static=False, per_channel=False
        )
        dynamic_quantizer.quantize(
            save_dir=self.onnx_path,
            quantization_config=dqconfig,
            file_suffix=f"quantized_{self.device}",
        )

    def load_quantized_model(self):
        if not self.use_onnx:
            raise ValueError(
                "Model must be in ONNX format to load quantized version. Set use_onnx=True when initializing."
            )
        self.model = ORTModelForFeatureExtraction.from_pretrained(
            self.onnx_path,
            file_name=f"model_optimized_quantized_{self.device}.onnx",
            device=self.device,
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.onnx_path)
        self.create_pipeline()

    def print_model_sizes(self):
        if not self.use_onnx:
            raise ValueError(
                "Model must be in ONNX format to print sizes. Set use_onnx=True when initializing."
            )
        size = (self.onnx_path / "model.onnx").stat().st_size / (1024 * 1024)
        optimized_size = (self.onnx_path / "model_optimized.onnx").stat().st_size / (
            1024 * 1024
        )
        quantized_size = (
            self.onnx_path / "model_optimized_quantized.onnx"
        ).stat().st_size / (1024 * 1024)
        print(f"Model file size: {size:.2f} MB")
        print(f"Optimized Model file size: {optimized_size:.2f} MB")
        print(f"Quantized Model file size: {quantized_size:.2f} MB")



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gradio


class TaskConfigMixin:
    def _set_text_classification_config(self, inputs):
        if self.task in [
            "text-classification",
            "text-classification-multi-label",
            "chat-classification",
            "chat-classification-multi-label",
            "image-classification",
            "image-classification-multi-label",
        ]:
            with gradio.Tab("Label selector"):
                with gradio.Column():
                    label_selector = gradio.Dropdown(
                        choices=[],
                        label="label",
                        allow_custom_value=True,
                        multiselect=True,
                    )

                def update_labels(_label_selector):
                    self.labels = _label_selector
                    _kwargs = {
                        "choices": _label_selector,
                        "label": "label",
                    }
                    return (
                        gradio.CheckboxGroup(**_kwargs)
                        if "multi-label" in self.task
                        else gradio.Radio(**_kwargs)
                    )

                def get_label_from_dataframe(_input_data_component, _label_selector):
                    if "suggestion" in _input_data_component.columns:
                        unique_labels = (
                            _input_data_component["suggestion"].unique().tolist()
                        )
                    else:
                        unique_labels = _label_selector
                    if unique_labels is None:
                        return gradio.Dropdown(
                            choices=[],
                            label="label",
                            allow_custom_value=True,
                            multiselect=True,
                        )
                    else:
                        labels = [str(label) for label in unique_labels]
                        labels = sorted(labels)
                        return gradio.Dropdown(
                            choices=labels,
                            value=labels,
                            label="label",
                            allow_custom_value=True,
                            multiselect=True,
                        )

                self.input_data_component.change(
                    fn=get_label_from_dataframe,
                    inputs=[self.input_data_component, label_selector],
                    outputs=[label_selector],
                )

                label_selector.change(
                    fn=update_labels,
                    inputs=[label_selector],
                    outputs=[
                        input
                        for input in inputs
                        if isinstance(input, (gradio.Radio, gradio.CheckboxGroup))
                    ],
                )



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argilla as rg


class ArgillaMixin:
    def get_argilla_dataset(self):
        class MockClient:
            a = "dataset"

        client = MockClient()
        client.api.datasets = "mock"
        return rg.Dataset(
            name="fake-dataset",
            settings=self._get_argilla_settings(),
        )

    def _get_argilla_settings(self):
        if self.task == "text-classification":
            return rg.Settings(
                fields=[rg.TextField(name="text")],
                questions=[rg.LabelQuestion(name="label", labels=self.labels)],
            )
        elif self.task == "text-classification-multi-label":
            return rg.Settings(
                fields=[rg.TextField(name="text")],
                questions=[rg.MultiLabelQuestion(name="label", labels=self.labels)],
            )
        elif self.task == "token-classification":
            raise NotImplementedError
        elif self.task == "question-answering":
            raise NotImplementedError
        elif self.task == "text-generation":
            return rg.Settings(
                fields=[rg.TextField(name="prompt")],
                questions=[rg.TextQuestion(name="completion")],
            )
        elif self.task == "text-generation-preference":
            return rg.Settings(
                fields=[
                    rg.TextField(name="prompt"),
                    rg.TextField(name="chosen"),
                    rg.TextField(name="rejected"),
                ],
                questions=[
                    rg.LabelQuestion(name="flag", labels=["A", "B", "tie"]),
                    rg.TextField(name="reason", required=False),
                ],
            )
        elif self.task == "chat-classification":
            raise NotImplementedError
        elif self.task == "chat-classification-multi-label":
            raise NotImplementedError
        elif self.task == "chat-generation":
            raise NotImplementedError
        elif self.task == "chat-generation-preference":
            raise NotImplementedError
        elif self.task == "image-classification":
            raise NotImplementedError
        elif self.task == "image-classification-multi-label":
            raise NotImplementedError
        elif self.task == "image-generation":
            raise NotImplementedError
        elif self.task == "image-description":
            raise NotImplementedError
        elif self.task == "image-generation-preference":
            raise NotImplementedError
        elif self.task == "image-question-answering":
            raise NotImplementedError
        else:
            raise NotImplementedError



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import warnings
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Literal, Optional, Union

import dash
import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
import umap
from dash import dash_table, dcc, html
from dash.dependencies import Input, Output, State
from plotly.graph_objs._figure import Figure

from dataset_viber._constants import DEFAULT_EMBEDDING_MODEL

if TYPE_CHECKING:
    from sentence_transformers import SentenceTransformer


class BulkInterface:
    def __init__(
        self,
        dataframe: pd.DataFrame,
        content_column: str,
        *,
        label_column: str = None,
        embedding_model: Optional[
            Union["SentenceTransformer", str]
        ] = DEFAULT_EMBEDDING_MODEL,
        umap_kwargs: dict = {},
        labels: list[str] = None,
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
        content_format: Optional[str] = "text",
    ):
        self.content_format = content_format
        self.content_column = content_column
        self.label_column = label_column
        self.labels = labels
        if label_column and labels:
            if not all([label in dataframe[label_column].unique() for label in labels]):
                # apply label to first x empty rows in label_column
                empty_rows = dataframe[dataframe[label_column] == ""].index
                for label, row in zip(labels, empty_rows):
                    dataframe.loc[row, label_column] = label
                warnings.warn(
                    "Labels were not found in the label_column. Applied labels to the first x empty rows."
                )

        contents = dataframe[content_column].tolist()

        # Apply embedding reduction
        component_columns: list[str] = ["x", "y"]
        if all([col in dataframe.columns for col in component_columns]):
            umap_df = dataframe[component_columns]
        else:
            if "embeddings" in dataframe.columns:
                embeddings = dataframe["embeddings"].tolist()
            else:
                self._set_embedding_model(embedding_model)
                embeddings = self.embed_content(contents)
            reducer = umap.UMAP(n_components=2, **umap_kwargs)
            umap_embeddings = reducer.fit_transform(embeddings)
            # Create a DataFrame for plotting
            umap_df = pd.DataFrame(
                umap_embeddings,
                columns=component_columns,
            )

        umap_df["index"] = dataframe.index
        for col in dataframe.columns:
            umap_df[col] = dataframe[col]

        self.umap_df = umap_df
        app = dash.Dash(__name__, external_stylesheets=[dbc.themes.SANDSTONE])
        figure = self._get_initial_figure(umap_df)
        app.layout = self._get_app_layout(figure, umap_df, labels, hf_token)

        if labels is not None:

            @app.callback(
                [
                    Output("scatter-plot", "figure", allow_duplicate=True),
                    Output("data-table", "data", allow_duplicate=True),
                    Output("data-table", "tooltip_data", allow_duplicate=True),
                ],
                [Input("update-button", "n_clicks")],
                [
                    State("scatter-plot", "selectedData"),
                    State("label-dropdown", "value"),
                    State("scatter-plot", "figure"),
                ],
                prevent_initial_call=True,
            )
            def update_labels(n_clicks, selectedData, new_label, current_figure):
                ctx = dash.callback_context
                if not ctx.triggered or new_label is None:
                    return current_figure, self.umap_df.to_dict("records")

                hidden_traces = []
                for trace in figure["data"]:
                    if trace["visible"] == "legendonly":
                        hidden_traces.append(trace)

                if selectedData and selectedData["points"]:
                    selected_indices = [
                        point["customdata"][0] for point in selectedData["points"]
                    ]
                    self.umap_df.loc[selected_indices, self.label_column] = new_label
                    updated_traces = []
                    points_to_move = defaultdict(list)
                    for trace in current_figure["data"]:
                        if trace not in hidden_traces:
                            if new_label != trace["name"]:
                                points_to_keep = defaultdict(list)
                                for idx, point in enumerate(trace["customdata"]):
                                    if point[0] not in selected_indices:
                                        points_to_keep["customdata"].append(point)
                                        points_to_keep["x"].append(trace["x"][idx])
                                        points_to_keep["y"].append(trace["y"][idx])
                                    else:
                                        points_to_move["customdata"].append(point)
                                        points_to_move["x"].append(trace["x"][idx])
                                        points_to_move["y"].append(trace["y"][idx])
                                trace["customdata"] = points_to_keep["customdata"]
                                trace["x"] = points_to_keep["x"]
                                trace["y"] = points_to_keep["y"]
                                trace["selectedpoints"] = []
                                updated_traces.append(trace)
                    for trace in current_figure["data"]:
                        if trace["name"] == new_label:
                            trace["customdata"] += points_to_move["customdata"]
                            trace["x"] += points_to_move["x"]
                            trace["y"] += points_to_move["y"]
                            trace["selectedpoints"] = []
                            updated_traces.append(trace)
                    current_figure["data"] = updated_traces

                local_dataframe = self.umap_df.copy()
                tooltip_data = self.get_tooltip(local_dataframe)
                if self.content_format == "chat":
                    local_dataframe[self.content_column] = local_dataframe[
                        self.content_column
                    ].apply(lambda x: x[0]["content"])
                return current_figure, local_dataframe.to_dict("records"), tooltip_data

            # Callback to print the dataframe
            @app.callback(
                Output("data-table", "data", allow_duplicate=True),
                [Input("upload-button", "n_clicks")],
                prevent_initial_call=True,
            )
            def print_dataframe(n_clicks):
                if n_clicks > 0:
                    print(self.umap_df)  # This will print the dataframe to the console
                return self.umap_df.to_dict(
                    "records"
                )  # Return the data to avoid updating the table

            @app.callback(
                Output("download-text", "data"),
                Input("btn-download-txt", "n_clicks"),
                prevent_initial_call=True,
            )
            def func(n_clicks):
                return dcc.send_data_frame(self.umap_df.to_csv, "data.csv")

        # Update the existing update_selection callback
        @app.callback(
            [
                Output("scatter-plot", "figure", allow_duplicate=True),
                Output("data-table", "data", allow_duplicate=True),
                Output("data-table", "tooltip_data", allow_duplicate=True),
            ],
            [Input("scatter-plot", "selectedData")],
            [State("scatter-plot", "figure")],
            prevent_initial_call=True,
        )
        def update_selection(selectedData, figure):
            ctx = dash.callback_context
            if not ctx.triggered:
                return figure, self.umap_df.to_dict("records")

            hidden_traces = []
            for trace in figure["data"]:
                if trace.get("visible") == "legendonly":
                    hidden_traces.append(trace)

            if selectedData and selectedData["points"]:
                selected_indices = [
                    point["customdata"][0] for point in selectedData["points"]
                ]
                filtered_df = self.umap_df.iloc[selected_indices]
            else:
                filtered_df = self.umap_df
                selected_indices = None

            if hidden_traces:
                filtered_df = filtered_df[
                    ~filtered_df[self.label_column].isin(hidden_traces)
                ]

            local_dataframe = filtered_df.copy()
            tooltip_data = self.get_tooltip(local_dataframe)
            if self.content_format == "chat":
                local_dataframe[self.content_column] = local_dataframe[
                    self.content_column
                ].apply(lambda x: x[0]["content"])
            return figure, local_dataframe.to_dict("records"), tooltip_data

        self.app = app

    def _set_embedding_model(self, embedding_model: str):
        import torch
        from sentence_transformers import SentenceTransformer

        if isinstance(embedding_model, SentenceTransformer):
            self.embedding_model = embedding_model
        elif isinstance(embedding_model, str):
            device = "cpu"
            if torch.backends.mps.is_available():
                device = "mps"
            elif torch.cuda.is_available():
                device = "cuda"
            if importlib.util.find_spec("onnxruntime") is not None:
                from dataset_viber.embedder import Embedder

                self.embedding_model = Embedder(model_id=embedding_model, device=device)
            else:
                self.embedding_model = SentenceTransformer(
                    model_name_or_path=embedding_model, device=device
                )
        else:
            raise ValueError(
                "Embedding model should be of type `str` or `SentenceTransformer`"
            )

    @classmethod
    def for_text_visualization(
        cls,
        dataframe: pd.DataFrame,
        content_column: str,
        *,
        label_column: str = None,
        embedding_model: Optional[
            Union["SentenceTransformer", str]
        ] = DEFAULT_EMBEDDING_MODEL,
        umap_kwargs: dict = {},
    ):
        return cls(
            dataframe=dataframe,
            content_column=content_column,
            label_column=label_column,
            embedding_model=embedding_model,
            umap_kwargs=umap_kwargs,
            content_format="text",
        )

    @classmethod
    def for_text_classification(
        cls,
        dataframe: pd.DataFrame,
        content_column: str,
        labels: list[str],
        *,
        label_column: str = None,
        embedding_model: Optional[
            Union["SentenceTransformer", str]
        ] = DEFAULT_EMBEDDING_MODEL,
        umap_kwargs: dict = {},
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ):
        if not label_column:
            dataframe["label"] = ""
            label_column = "label"
        return cls(
            dataframe=dataframe,
            content_column=content_column,
            label_column=label_column,
            embedding_model=embedding_model,
            umap_kwargs=umap_kwargs,
            labels=labels,
            dataset_name=dataset_name,
            hf_token=hf_token,
            private=private,
            content_format="text",
        )

    @classmethod
    def for_chat_visualization(
        cls,
        dataframe: pd.DataFrame,
        chat_column: List[Dict[str, str]],
        *,
        label_column: str = None,
        embedding_model: Optional[
            Union["SentenceTransformer", str]
        ] = DEFAULT_EMBEDDING_MODEL,
        umap_kwargs: dict = {},
    ):
        return cls(
            dataframe=dataframe,
            content_column=chat_column,
            label_column=label_column,
            embedding_model=embedding_model,
            umap_kwargs=umap_kwargs,
            content_format="chat",
        )

    @classmethod
    def for_chat_classification(
        cls,
        dataframe: pd.DataFrame,
        chat_column: List[Dict[str, str]],
        labels: list[str],
        *,
        label_column: str = None,
        embedding_model: Optional[
            Union["SentenceTransformer", str]
        ] = DEFAULT_EMBEDDING_MODEL,
        umap_kwargs: dict = {},
        dataset_name: Optional[str] = None,
        hf_token: Optional[str] = None,
        private: Optional[bool] = False,
    ):
        if not label_column:
            dataframe["label"] = ""
            label_column = "label"
        return cls(
            dataframe=dataframe,
            content_column=chat_column,
            embedding_model=embedding_model,
            umap_kwargs=umap_kwargs,
            labels=labels,
            dataset_name=dataset_name,
            label_column=label_column,
            hf_token=hf_token,
            private=private,
            content_format="chat",
        )

    def launch(self, **kwargs):
        self.app.run_server(**kwargs)

    def _get_initial_figure(self, dataframe) -> Figure:
        # color_map = {label: color for label, color in zip(self.labels, _COLORS)}
        dataframe[f"wrapped_hover_{self.content_column}"] = dataframe[
            self.content_column
        ].apply(lambda x: self.format_content(x, content_format=self.content_format))
        custom_data = ["index"] + [
            col
            for col in dataframe.columns
            if col not in ["x", "y", "index", self.content_column]
        ]
        hovertemplate: Literal[""] = ""
        df_custom = dataframe[custom_data]
        for col in df_custom.columns:
            if col in ["index", self.label_column]:
                continue
            idx = df_custom.columns.get_loc(col)
            hovertemplate += f"<b>{col.replace('wrapped_hover_', '')}</b>:<br>%{{customdata[{idx}]}}<br>"
        fig = px.scatter(
            dataframe,
            x="x",
            y="y",
            color=self.label_column if self.label_column in dataframe.columns else None,
            height=800,
            custom_data=custom_data,
        )
        fig.update_traces(hovertemplate=str(hovertemplate))
        fig.update_layout(
            xaxis_title=None,
            yaxis_title=None,
            xaxis=dict(showticklabels=False),
            yaxis=dict(showticklabels=False),
            margin=dict(l=0, r=0, t=0, b=0),
            dragmode="lasso",
            legend=dict(
                orientation="h",  # Horizontal legend
                yanchor="top",  # Anchor the legend to the top of the container
                y=-0.01,  # Position the legend below the plot
                xanchor="center",  # Center the legend horizontally
                x=0.5,  # Center the legend at the bottom
            ),
            hoverlabel=dict(
                font_size=10,
                font_family="monospace",
            ),
        )
        return fig

    def _get_app_layout(self, figure, dataframe, labels, hf_token):
        local_dataframe = dataframe.copy()
        buttons = []
        if labels is not None:
            buttons.extend(
                [
                    buttons.append(
                        dcc.Dropdown(
                            id="label-dropdown",
                            options=[
                                {"label": label, "value": label} for label in labels
                            ],
                            value=labels[0],
                            clearable=True,
                            style={
                                "width": "200px",
                                "marginBottom": "-13px",
                                "display": "inline-block",
                            },
                        )
                    ),
                    buttons.append(
                        dbc.Button(
                            "Update Labels",
                            id="update-button",
                            n_clicks=0,
                        )
                    ),
                    dbc.Button(
                        "Upload to Hub",
                        id="upload-button",
                        n_clicks=0,
                    ),
                    dbc.Button("Download Text", id="btn-download-txt"),
                    dcc.Download(id="download-text"),
                ]
            )
        if self.content_format == "chat":
            tooltip_data = self.get_tooltip(local_dataframe)
            local_dataframe[self.content_column] = local_dataframe[
                self.content_column
            ].apply(lambda x: x[0]["content"])
            columns = local_dataframe.columns
        elif self.content_format == "text":
            tooltip_data = None
            columns = local_dataframe.columns
        else:
            raise ValueError(
                "content_format should be either 'text' or 'chat' but got {self.content_format}"
            )

        layout = html.Div(
            [
                html.H1("BulkInterface"),
                # Scatter plot
                html.Div(
                    [
                        dcc.Graph(id="scatter-plot", figure=figure),
                        html.Div([*buttons]),
                    ],
                    style={
                        "width": "49%",
                        "display": "inline-block",
                        "vertical-align": "top",
                        "marginRight": "1%",
                    },
                ),
                html.Div(
                    [
                        dash_table.DataTable(
                            id="data-table",
                            columns=[{"name": i, "id": i} for i in columns],
                            data=local_dataframe[columns].to_dict("records"),
                            hidden_columns=[
                                "x",
                                "y",
                                "index",
                                f"wrapped_hover_{self.content_column}",
                            ],
                            tooltip_data=tooltip_data,
                            column_selectable=False,
                            page_size=20,
                            fill_width=True,
                            css=[
                                {"selector": ".show-hide", "rule": "display: none"},
                                {
                                    "selector": ".dash-table-tooltip",
                                    "rule": """
                                        background-color: grey;
                                        font-family: monospace;
                                        color: white;
                                        max-width: 100vw !important;
                                        max-height: 80vh !important;
                                        overflow: auto;
                                        font-size: 10px;
                                        position: fixed;
                                        top: 50%;
                                        left: 50%;
                                        transform: translate(-50%, -50%);
                                        z-index: 1000;
                                    """,
                                },
                            ],
                            style_cell={
                                "whiteSpace": "normal",
                                "height": "auto",
                                "textAlign": "left",
                                "font-size": "10px",
                                "overflow": "auto",  # Enable scrolling
                            },
                            style_data={
                                "whiteSpace": "normal",
                                "height": "auto",
                            },
                            style_table={"overflowX": "auto"},
                            tooltip_duration=None,
                        )
                    ],
                    style={
                        "width": "50%",
                        "display": "inline-block",
                        "vertical-align": "top",
                    },
                ),
            ]
        )
        return layout

    def embed_content(self, content: List[str]):
        if self.content_format == "text":
            return self.embedding_model.encode(content, convert_to_numpy=True)
        elif self.content_format == "chat":
            content = [
                " ".join([turn["content"] for turn in conversation])
                for conversation in content
            ]
            return self.embedding_model.encode(content, convert_to_numpy=True)

    def format_content(self, content, max_length=120, content_format="text"):
        wrapped_text = ""
        if content_format == "text":
            words = content.split(" ")
            line = ""

            for word in words:
                if len(line) + len(word) + 1 > max_length:
                    if line:
                        wrapped_text += line + "<br>"
                    line = word
                else:
                    if line:
                        line += " " + word
                    else:
                        line = word

            wrapped_text += line
            return wrapped_text
        elif content_format == "chat":
            wrapped_text = "First 2 turns:<br><br>"
            for turn in content[:3]:
                wrapped_text += f"<b>{turn['role']}</b>:<br>{self.format_content(turn['content'])}<br><br>"
            return wrapped_text

    def get_tooltip(self, dataframe):
        if self.content_format == "text":
            return None
        return [
            {
                self.content_column: {
                    "value": pd.DataFrame.from_records(value)[
                        ["role", "content"]
                    ].to_markdown(index=False, tablefmt="pipe"),
                    "type": "markdown",
                }
            }
            for value in dataframe[self.content_column].tolist()
        ]



# Copyright 2024-present, David Berenstein, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import random
import time

import requests
from PIL import Image

from dataset_viber import AnnotatorInterFace

HF_TOKEN = os.environ["HF_TOKEN"]
HEADERS = {"Authorization": f"Bearer {HF_TOKEN}"}
DATASET_SERVER_URL = "https://datasets-server.huggingface.co"
DATASET_NAME = "poloclub%2Fdiffusiondb&config=2m_random_1k&split=train"
MODEL_URL = (
    "https://api-inference.huggingface.co/models/black-forest-labs/FLUX.1-schnell"
)


def retrieve_sample(idx):
    api_url = f"{DATASET_SERVER_URL}/rows?dataset={DATASET_NAME}&offset={idx}&length=1"
    response = requests.get(api_url, headers=HEADERS)
    data = response.json()
    img_url = data["rows"][0]["row"]["image"]["src"]
    prompt = data["rows"][0]["row"]["prompt"]
    return img_url, prompt


def get_rows():
    api_url = f"{DATASET_SERVER_URL}/size?dataset={DATASET_NAME}"
    response = requests.get(api_url, headers=HEADERS)
    num_rows = response.json()["size"]["config"]["num_rows"]
    return num_rows


def generate_response(prompt):
    def _get_response(prompt):
        payload = {
            "inputs": prompt,
        }
        response = requests.post(MODEL_URL, headers=HEADERS, json=payload)
        if response.status_code != 200:
            time.sleep(10)
            return _get_response(prompt)
        return response

    response = _get_response(prompt)
    image = Image.open(io.BytesIO(response.content))
    return image


def next_input(_prompt, _completion_a, _completion_b):
    random_idx = random.randint(0, get_rows()) - 1
    img_url, prompt = retrieve_sample(random_idx)
    generated_image = generate_response(prompt)
    return (prompt, img_url, generated_image)


if __name__ == "__main__":
    interface = AnnotatorInterFace.for_image_generation_preference(
        interactive=False, fn_next_input=next_input
    )
    interface.launch()
